@@ -10,12 +10,12 @@
 #endif
 
 
-#if HAVE_SYS_UTSNAME_H
+#ifdef HAVE_SYS_UTSNAME_H
 #include <sys/utsname.h>
 #endif
 #include <ctype.h>	/* XXX for /etc/rpm/platform contents */
 
-#if HAVE_SYS_SYSTEMCFG_H
+#ifdef HAVE_SYS_SYSTEMCFG_H
 #include <sys/systemcfg.h>
 #else
 #define __power_pc() 0
@@ -32,9 +32,11 @@
 #include <rpm/rpmlog.h>
 #include <rpm/argv.h>
 
-#include "rpmio/rpmlua.h"
-#include "rpmio/rpmio_internal.h"	/* XXX for rpmioSlurp */
-#include "lib/misc.h"
+#include "rpmlua.h"
+#include "rpmio_internal.h"	/* XXX for rpmioSlurp */
+#include "misc.h"
+#include "backend/dbi.h"
+#include "rpmug.h"
 
 #include "debug.h"
 
@@ -71,7 +73,7 @@ struct rpmvarValue * next;
 };
 
 struct rpmOption {
-    char * name;
+    const char * name;
     int var;
     int archSpecific;
     int macroize;
@@ -99,7 +101,7 @@ typedef struct canonEntry_s {
  * for giggles, 'key'_canon, 'key'_compat, and 'key'_canon will also work
  */
 typedef struct tableType_s {
-    char * const key;
+    const char *key;
     const int hasCanon;
     const int hasTranslate;
     struct machEquivTable_s equiv;
@@ -160,7 +162,6 @@ static void rpmRebuildTargetVars(rpmrcCtx ctx, const char **target, const char *
 static rpmrcCtx rpmrcCtxAcquire(int write)
 {
     static struct rpmrcCtx_s _globalCtx = {
-	.lock = PTHREAD_RWLOCK_INITIALIZER,
 	.currTables = { RPM_MACHTABLE_INSTOS, RPM_MACHTABLE_INSTARCH },
 	.tables = {
 	    { "arch", 1, 0 },
@@ -168,6 +169,7 @@ static rpmrcCtx rpmrcCtxAcquire(int write)
 	    { "buildarch", 0, 1 },
 	    { "buildos", 0, 1 }
 	},
+	.lock = PTHREAD_RWLOCK_INITIALIZER,
     };
     rpmrcCtx ctx = &_globalCtx;
 
@@ -263,7 +265,7 @@ static int machCompatCacheAdd(char * name, const char * fn, int linenum,
 	    entry->equivs = xrealloc(entry->equivs, sizeof(*entry->equivs)
 					* (entry->count + 1));
 	else
-	    entry->equivs = xmalloc(sizeof(*entry->equivs));
+	    entry->equivs = (char **)xmalloc(sizeof(*entry->equivs));
 
 	entry->equivs[entry->count] = xstrdup(chptr);
 	entry->count++;
@@ -295,7 +297,7 @@ static void machAddEquiv(machEquivTable table, const char * name,
 	    table->list = xrealloc(table->list, (table->count + 1)
 				    * sizeof(*table->list));
 	else
-	    table->list = xmalloc(sizeof(*table->list));
+	    table->list = (machEquivInfo)xmalloc(sizeof(*table->list));
 
 	table->list[table->count].name = xstrdup(name);
 	table->list[table->count++].score = distance;
@@ -453,29 +455,56 @@ const char * lookupInDefaultTable(const char * name,
 
 static void setDefaults(void)
 {
+    /* If either is missing, we need to go through this whole dance */
+    if (defrcfiles && macrofiles)
+	return;
+
     const char *confdir = rpmConfigDir();
-    if (!defrcfiles) {
-	defrcfiles = rstrscat(NULL, confdir, "/rpmrc", ":",
-				confdir, "/" RPMCANONVENDOR "/rpmrc", ":",
-				SYSCONFDIR "/rpmrc", ":",
-			  	"~/.rpmrc", NULL);
+    const char *xdgconf = getenv("XDG_CONFIG_HOME");
+    if (!(xdgconf && *xdgconf))
+	xdgconf = "~/.config";
+    char *userdir = rpmGetPath(xdgconf, "/rpm", NULL);
+    char *usermacros = rpmGetPath(userdir, "/macros", NULL);
+    char *userrc = rpmGetPath(userdir, "/rpmrc", NULL);
+
+    /*
+     * Prefer XDG_CONFIG_HOME/rpm/{macros,rpmrc} but fall back to ~/.rpmmacros
+     * and ~/.rpmrc iff either exists and the XDG rpm directory doesn't.
+     */
+    if (rpmGlob(userdir, NULL, NULL)) {
+	const char *oldmacros = "~/.rpmmacros";
+	const char *oldrc = "~/.rpmrc";
+	if (rpmGlob(oldmacros, NULL, NULL) == 0 || rpmGlob(oldrc, NULL, NULL) == 0) {
+	    free(usermacros);
+	    free(userrc);
+	    usermacros = xstrdup(oldmacros);
+	    userrc = xstrdup(oldrc);
+	}
     }
 
-#ifndef MACROFILES
+    if (!defrcfiles) {
+	defrcfiles = rstrscat(NULL, confdir, "/rpmrc", ":",
+				confdir, "/" RPM_VENDOR "/rpmrc", ":",
+				SYSCONFDIR "/rpmrc", ":",
+				userrc, NULL);
+    }
+
+    /* macrofiles may be pre-set from --macros */
     if (!macrofiles) {
 	macrofiles = rstrscat(NULL, confdir, "/macros", ":",
 				confdir, "/macros.d/macros.*", ":",
 				confdir, "/platform/%{_target}/macros", ":",
 				confdir, "/fileattrs/*.attr", ":",
-  				confdir, "/" RPMCANONVENDOR "/macros", ":",
+				confdir, "/" RPM_VENDOR "/macros", ":",
 				SYSCONFDIR "/rpm/macros.*", ":",
 				SYSCONFDIR "/rpm/macros", ":",
 				SYSCONFDIR "/rpm/%{_target}/macros", ":",
-				"~/.rpmmacros", NULL);
+				usermacros, NULL);
     }
-#else
-    macrofiles = MACROFILES;
-#endif
+
+    free(usermacros);
+    free(userrc);
+    free(userdir);
 }
 
 /* FIX: se usage inconsistent, W2DO? */
@@ -528,8 +557,8 @@ static rpmRC doReadRC(rpmrcCtx ctx, const char * urlfn)
 
 	/* Find keyword in table */
 	searchOption.name = s;
-	option = bsearch(&searchOption, optionTable, optionTableSize,
-			 sizeof(optionTable[0]), optionCompare);
+	option = (struct rpmOption *)bsearch(&searchOption, optionTable,
+			optionTableSize, sizeof(optionTable[0]), optionCompare);
 
 	if (option) {	/* For configuration variables  ... */
 	    const char *arch, *val;
@@ -580,7 +609,7 @@ static rpmRC doReadRC(rpmrcCtx ctx, const char * urlfn)
 	    if (option->macroize &&
 	      (arch == NULL || rstreq(arch, ctx->current[ARCH]))) {
 		char *n, *name;
-		n = name = xmalloc(strlen(option->name)+2);
+		n = name = (char *)xmalloc(strlen(option->name)+2);
 		if (option->localize)
 		    *n++ = '_';
 		strcpy(n, option->name);
@@ -733,6 +762,98 @@ exit:
     return rc;
 }
 
+#	if defined(__linux__) && defined(__x86_64__)
+#           if defined(__has_builtin)
+#               if __has_builtin(__builtin_cpu_supports)
+#                   define HAS_BUILTIN_CPU_SUPPORTS
+#               endif
+#           endif
+#           if defined(HAS_BUILTIN_CPU_SUPPORTS)
+static inline void cpuid(uint32_t op, uint32_t op2, uint32_t *eax, uint32_t *ebx, uint32_t *ecx, uint32_t *edx)
+{
+    asm volatile (
+	"cpuid\n"
+    : "=a" (*eax), "=b" (*ebx), "=c" (*ecx), "=d" (*edx)
+    : "a" (op), "c" (op2));
+}
+
+/* From gcc's gcc/config/i386/cpuid.h */
+/* Features (%eax == 1) */
+/* %ecx */
+#define bit_SSE3	(1 << 0)
+#define bit_SSSE3	(1 << 9)
+#define bit_FMA		(1 << 12)
+#define bit_CMPXCHG16B	(1 << 13)
+#define bit_SSE4_1	(1 << 19)
+#define bit_SSE4_2	(1 << 20)
+#define bit_MOVBE	(1 << 22)
+#define bit_POPCNT	(1 << 23)
+#define bit_OSXSAVE	(1 << 27)
+#define bit_AVX		(1 << 28)
+#define bit_F16C	(1 << 29)
+
+/* Extended Features (%eax == 0x80000001) */
+/* %ecx */
+#define bit_LAHF_LM	(1 << 0)
+#define bit_LZCNT	(1 << 5)
+
+/* Extended Features (%eax == 7) */
+/* %ebx */
+#define bit_BMI		(1 << 3)
+#define bit_AVX2	(1 << 5)
+#define bit_BMI2	(1 << 8)
+#define bit_AVX512F	(1 << 16)
+#define bit_AVX512DQ	(1 << 17)
+#define bit_AVX512CD	(1 << 28)
+#define bit_AVX512BW	(1 << 30)
+#define bit_AVX512VL	(1u << 31)
+
+static int get_x86_64_level(void)
+{
+    int level = 1;
+
+    unsigned int op_0_eax = 0, op_1_ecx = 0, op_80000001_ecx = 0, op_7_ebx = 0, unused;
+    cpuid(0, 0, &op_0_eax, &unused, &unused, &unused);
+    cpuid(1, 0, &unused, &unused, &op_1_ecx, &unused);
+    cpuid(0x80000001, 0, &unused, &unused, &op_80000001_ecx, &unused);
+    cpuid(7, 0, &unused, &op_7_ebx, &unused, &unused);
+
+    /* CPUID is unfortunately not quite enough: It indicates whether the hardware supports it,
+     * but software support (kernel/hypervisor) is also needed for register saving/restoring.
+     * For that we can ask the compiler's runtime lib through __builtin_cpu_supports. This is
+     * not usable as a complete replacement for the CPUID code though as not all extensions are
+     * supported (yet). */
+
+    const unsigned int op_1_ecx_lv2 = bit_SSE3 | bit_SSSE3 | bit_CMPXCHG16B | bit_SSE4_1 | bit_SSE4_2 | bit_POPCNT;
+    if ((op_1_ecx & op_1_ecx_lv2) == op_1_ecx_lv2 && (op_80000001_ecx & bit_LAHF_LM))
+	level = 2;
+
+    const unsigned int op_1_ecx_lv3 = bit_FMA | bit_MOVBE | bit_OSXSAVE | bit_AVX | bit_F16C;
+    const unsigned int op_7_ebx_lv3 = bit_BMI | bit_AVX2 | bit_BMI2;
+    if (level == 2 && (op_1_ecx & op_1_ecx_lv3) == op_1_ecx_lv3
+        && op_0_eax >= 7 && (op_7_ebx & op_7_ebx_lv3) == op_7_ebx_lv3
+        && (op_80000001_ecx & bit_LZCNT)
+        && __builtin_cpu_supports("avx2"))
+    {
+        level = 3;
+    }
+
+    const unsigned int op_7_ebx_lv4 = bit_AVX512F | bit_AVX512DQ | bit_AVX512CD | bit_AVX512BW | bit_AVX512VL;
+    if (level == 3 && (op_7_ebx & op_7_ebx_lv4) == op_7_ebx_lv4
+        && __builtin_cpu_supports("avx512f"))
+    {
+        level = 4;
+    }
+
+    return level;
+}
+#           else /* defined(HAS_BUILTIN_CPU_SUPPORTS) */
+static int get_x86_64_level(void)
+{
+    return 1;
+}
+#           endif
+#	endif
 
 #	if defined(__linux__) && defined(__i386__)
 #include <setjmp.h>
@@ -815,8 +936,8 @@ static inline int RPMClass(void)
 	cpu = (tfms>>8)&15;
 	
 	if (cpu == 5
-	    && cpuid_ecx(0) == '68xM'
-	    && cpuid_edx(0) == 'Teni'
+	    && cpuid_ecx(0) == 0x3638784d /* '68xM' */
+	    && cpuid_edx(0) == 0x54656e69 /* 'Teni' */
 	    && (cpuid_edx(1) & ((1<<8)|(1<<15))) == ((1<<8)|(1<<15))) {
 		sigaction(SIGILL, &oldsa, NULL);
 		return 6;	/* has CX8 and CMOV */
@@ -866,6 +987,31 @@ static int is_athlon(void)
 	return 1;
 }
 
+static unsigned int x86_family(unsigned int processor_sig)
+{
+    unsigned int base_family = (processor_sig >> 8) & 0x0f;
+    unsigned int family;
+    if (base_family == 0x0f) {
+	unsigned int extended_family = (processor_sig >> 20) & 0x0ff;
+	family = base_family + extended_family;
+    } else
+	family = base_family;
+    return family;
+}
+
+static unsigned int x86_model(unsigned int processor_sig)
+{
+    unsigned int base_model = (processor_sig >> 4) & 0x0f;
+    unsigned int family = x86_family(processor_sig);
+    unsigned int model;
+    if (family >= 0x6) {
+	unsigned int extended_model = (processor_sig >> 16) & 0x0f;
+	model = (extended_model << 4) | base_model;
+    } else
+	model = base_model;
+    return model;
+}
+
 static int is_pentium3(void)
 {
     unsigned int eax, ebx, ecx, edx, family, model;
@@ -878,8 +1024,8 @@ static int is_pentium3(void)
     if (!rstreqn(vendor, "GenuineIntel", 12))
 	return 0;
     cpuid(1, &eax, &ebx, &ecx, &edx);
-    family = (eax >> 8) & 0x0f;
-    model = (eax >> 4) & 0x0f;
+    family = x86_family(eax);
+    model = x86_model(eax);
     if (family == 6)
 	switch (model)
 	{
@@ -905,8 +1051,8 @@ static int is_pentium4(void)
     if (!rstreqn(vendor, "GenuineIntel", 12))
 	return 0;
     cpuid(1, &eax, &ebx, &ecx, &edx);
-    family = (eax >> 8) & 0x0f;
-    model = (eax >> 4) & 0x0f;
+    family = x86_family(eax);
+    model = x86_model(eax);
     if (family == 15)
 	switch (model)
 	{
@@ -935,8 +1081,8 @@ static int is_geode(void)
     if (!rstreqn(vendor, "AuthenticAMD", 12))
         return 0;
     cpuid(1, &eax, &ebx, &ecx, &edx);
-    family = (eax >> 8) & 0x0f;
-    model = (eax >> 4) & 0x0f;
+    family = x86_family(eax);
+    model = x86_model(eax);
     if (family == 5)
 	switch (model)
 	{
@@ -1039,6 +1185,8 @@ static void defaultMachine(rpmrcCtx ctx, const char ** arch, const char ** os)
 	    strcpy(un.machine, "i386");
 #elif defined(__x86_64__)
 	    strcpy(un.machine, "x86_64");
+#elif defined(__aarch64__)
+	    strcpy(un.machine, "aarch64");
 #else
 	    #warning "No architecture defined! Automatic detection may not work!"
 #endif 
@@ -1110,6 +1258,11 @@ static void defaultMachine(rpmrcCtx ctx, const char ** arch, const char ** os)
 #		endif
 #	endif
 
+#if defined(__loongarch64)
+	strcpy(un.machine, "loongarch64");
+#endif
+
+
 #if defined(__linux__)
 	/* in linux, lets rename parisc to hppa */
 	if (rstreq(un.machine, "parisc"))
@@ -1283,6 +1436,16 @@ static void defaultMachine(rpmrcCtx ctx, const char ** arch, const char ** os)
 	}
 #	endif
 
+# if defined(__linux__) && defined(__x86_64__)
+	{
+	    int x86_64_level = get_x86_64_level();
+	    if (x86_64_level > 1) {
+	        strcpy(un.machine, "x86_64_vX");
+	        un.machine[8] = '0' + x86_64_level;
+	    }
+	}
+#endif
+
 	/* the uname() result goes through the arch_canon table */
 	canon = lookupInCanonTable(un.machine,
 			   ctx->tables[RPM_MACHTABLE_INSTARCH].canons,
@@ -1346,7 +1509,7 @@ static void rpmSetVarArch(rpmrcCtx ctx,
 	    next->value = _free(next->value);
 	    next->arch = _free(next->arch);
 	} else if (next->arch || arch) {
-	    next->next = xmalloc(sizeof(*next->next));
+	    next->next = (struct rpmvarValue *)xmalloc(sizeof(*next->next));
 	    next = next->next;
 	    next->value = NULL;
 	    next->arch = NULL;
@@ -1550,6 +1713,14 @@ static void rpmRebuildTargetVars(rpmrcCtx ctx,
  * XXX Make sure that per-arch optflags is initialized correctly.
  */
   { const char *optflags = rpmGetVarArch(ctx, RPMVAR_OPTFLAGS, ca);
+    /*
+     * If not defined for the target arch, fall back to current arch
+     * definitions, with buildarchtranslate applied. This is WRONG
+     * but it's what we've always done, except for buildarchtranslate.
+     */
+    if (optflags == NULL) {
+	optflags = rpmGetVarArch(ctx, RPMVAR_OPTFLAGS, NULL);
+    }
     if (optflags != NULL) {
 	rpmPopMacro(NULL, "optflags");
 	rpmPushMacro(NULL, "optflags", NULL, optflags, RMIL_RPMRC);
@@ -1587,7 +1758,7 @@ static rpmRC rpmReadRC(rpmrcCtx ctx, const char * rcfiles)
     argvSplit(&globs, rcfiles, ":");
     for (p = globs; *p; p++) {
 	ARGV_t av = NULL;
-	if (rpmGlob(*p, NULL, &av) == 0) {
+	if (rpmGlobPath(*p, RPMGLOB_NOCHECK, NULL, &av) == 0) {
 	    argvAppend(&files, av);
 	    argvFree(av);
 	}
@@ -1615,22 +1786,13 @@ exit:
     return rc;
 }
 
-static void register_atexit(void)
-{
-    if (atexit(rpmAtExit) != 0)
-	rpmlog(RPMLOG_WARNING, _("failed to register exit handler"));
-}
-
 /* External interfaces */
 
 int rpmReadConfigFiles(const char * file, const char * target)
 {
-    static pthread_once_t atexit_registered = PTHREAD_ONCE_INIT;
     int rc = -1; /* assume failure */
     rpmrcCtx ctx = rpmrcCtxAcquire(1);
 
-    pthread_once(&atexit_registered, register_atexit);
-
     if (rpmInitCrypto())
 	goto exit;
 
@@ -1659,10 +1821,8 @@ int rpmReadConfigFiles(const char * file, const char * target)
 	free(os);
     }
 
-#ifdef WITH_LUA
     /* Force Lua state initialization */
     rpmluaGetGlobalState();
-#endif
     rc = 0;
 
 exit:
@@ -1738,10 +1898,9 @@ void rpmFreeRpmrc(void)
 
     /* XXX doesn't really belong here but... */
     rpmFreeCrypto();
-#ifdef WITH_LUA
     rpmlua lua = rpmluaGetGlobalState();
     rpmluaFree(lua);
-#endif
+    rpmugFree();
 
     rpmrcCtxRelease(ctx);
     return;
@@ -1792,6 +1951,8 @@ int rpmShowRC(FILE * fp)
 	fprintf(fp," %s", equivTable->list[i].name);
     fprintf(fp, "\n");
 
+    dbShowRC(fp);
+
     fprintf(fp, "\nRPMRC VALUES:\n");
     for (i = 0, opt = optionTable; i < optionTableSize; i++, opt++) {
 	const char *s = rpmGetVarArch(ctx, opt->var, NULL);
