@@ -4,6 +4,7 @@
 #include <sys/wait.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <signal.h>
 
 #include <rpm/rpmfileutil.h>
 #include <rpm/rpmmacro.h>
@@ -12,16 +13,17 @@
 #include <rpm/header.h>
 #include <rpm/rpmds.h>
 
-#include "rpmio/rpmlua.h"
-#include "lib/rpmscript.h"
-#include "rpmio/rpmio_internal.h"
+#include "rpmlua.h"
+#include "rpmscript.h"
+#include "rpmio_internal.h"
 
-#include "lib/rpmplugins.h"     /* rpm plugins hooks */
+#include "rpmchroot.h"
+#include "rpmplugins.h"     /* rpm plugins hooks */
 
 #include "debug.h"
 
 struct scriptNextFileFunc_s {
-    char *(*func)(void *);	/* function producing input for script */
+    nextfilefunc func;		/* function producing input for script */
     void *param;		/* parameter for func */
 };
 
@@ -34,7 +36,9 @@ struct rpmScript_s {
     char *body;			/* script body */
     char *descr;		/* description for logging */
     rpmscriptFlags flags;	/* flags to control operation */
-    struct scriptNextFileFunc_s nextFileFunc;  /* input function */
+    char *rpmver;		/* builder rpm version */
+    int chroot;			/* chrooted script? */
+    struct scriptNextFileFunc_s *nextFileFunc;  /* input function */
 };
 
 struct scriptInfo_s {
@@ -66,6 +70,12 @@ static const struct scriptInfo_s scriptInfo[] = {
     { RPMSCRIPT_POSTTRANS, "posttrans", 0,
 	RPMTAG_POSTTRANS, RPMTAG_POSTTRANSPROG, RPMTAG_POSTTRANSFLAGS,
 	0, },
+    { RPMSCRIPT_PREUNTRANS, "preuntrans", 0,
+	RPMTAG_PREUNTRANS, RPMTAG_PREUNTRANSPROG, RPMTAG_PREUNTRANSFLAGS,
+	RPMSCRIPT_FLAG_CRITICAL, },
+    { RPMSCRIPT_POSTUNTRANS, "postuntrans", 0,
+	RPMTAG_POSTUNTRANS, RPMTAG_POSTUNTRANSPROG, RPMTAG_POSTUNTRANSFLAGS,
+	0, },
     { RPMSCRIPT_TRIGGERPREIN, "triggerprein", RPMSENSE_TRIGGERPREIN,
 	RPMTAG_TRIGGERPREIN, 0, 0,
 	0, },
@@ -81,6 +91,9 @@ static const struct scriptInfo_s scriptInfo[] = {
     { RPMSCRIPT_VERIFY, "verify", 0,
 	RPMTAG_VERIFYSCRIPT, RPMTAG_VERIFYSCRIPTPROG, RPMTAG_VERIFYSCRIPTFLAGS,
 	RPMSCRIPT_FLAG_CRITICAL, },
+    { RPMSCRIPT_SYSUSERS, "sysusers", 0,
+	RPMTAG_SYSUSERS, 0, 0,
+	RPMSCRIPT_FLAG_CRITICAL, },
     { 0, "unknown", 0,
 	RPMTAG_NOT_FOUND, RPMTAG_NOT_FOUND, RPMTAG_NOT_FOUND,
 	0, }
@@ -93,60 +106,93 @@ static const struct scriptInfo_s * findTag(rpmTagVal tag)
 	si++;
     return si;
 }
+
+static int next_file(lua_State *L)
+{
+    scriptNextFileFunc nff = (scriptNextFileFunc)lua_touserdata(L, lua_upvalueindex(1));
+    lua_pushstring(L, nff->func(nff->param));
+    return 1;
+}
+
+int rpmScriptChrootIn(rpmScript script)
+{
+    int rc = 0;
+    if (script->chroot)
+	rc = rpmChrootIn();
+    return rc;
+}
+
+int rpmScriptChrootOut(rpmScript script)
+{
+    int rc = 0;
+    if (script->chroot)
+	rc = rpmChrootOut();
+    return rc;
+}
 /**
  * Run internal Lua script.
  */
 static rpmRC runLuaScript(rpmPlugins plugins, ARGV_const_t prefixes,
-		   const char *sname, rpmlogLvl lvl, FD_t scriptFd,
-		   ARGV_t * argvp, const char *script, int arg1, int arg2,
-		   scriptNextFileFunc nextFileFunc)
+		   rpmScript script, rpmlogLvl lvl, FD_t scriptFd,
+		   ARGV_t * argvp, int arg1, int arg2)
 {
+    char *scriptbuf = script->body;
     rpmRC rc = RPMRC_FAIL;
-#ifdef WITH_LUA
-    ARGV_t argv = argvp ? *argvp : NULL;
-    rpmlua lua = NULL; /* Global state. */
-    rpmluav var = rpmluavNew();
+    rpmlua lua = rpmluaGetGlobalState();
+    lua_State *L = (lua_State *)rpmluaGetLua(lua);
     int cwd = -1;
 
-    rpmlog(RPMLOG_DEBUG, "%s: running <lua> scriptlet.\n", sname);
+    rpmlog(RPMLOG_DEBUG, "%s: running <lua> scriptlet.\n", script->descr);
 
-    /* Create arg variable */
-    rpmluaPushTable(lua, "arg");
-    rpmluavSetListMode(var, 1);
-    rpmluaSetNextFileFunc(nextFileFunc->func, nextFileFunc->param);
-    if (argv) {
-	char **p;
-	for (p = argv; *p; p++) {
-	    rpmluavSetValue(var, RPMLUAV_STRING, *p);
-	    rpmluaSetVar(lua, var);
+    if (script->nextFileFunc) {
+	lua_getglobal(L, "rpm");
+	lua_pushlightuserdata(L, script->nextFileFunc);
+	lua_pushcclosure(L, &next_file, 1);
+	lua_setfield(L, -2, "next_file");
+    }
+
+    if (prefixes) {
+	lua_newtable(L);
+	for (ARGV_const_t p = prefixes; p && *p; p++) {
+	    int i = p - prefixes + 1;
+	    lua_pushstring(L, *p);
+	    lua_rawseti(L, -2, i);
 	}
+	lua_setglobal(L, "RPM_INSTALL_PREFIX");
     }
-    if (arg1 >= 0) {
-	rpmluavSetValueNum(var, arg1);
-	rpmluaSetVar(lua, var);
+
+    if (arg1 >= 0)
+	argvAddNum(argvp, arg1);
+    if (arg2 >= 0)
+	argvAddNum(argvp, arg2);
+
+    if (arg1 >= 0 || arg2 >= 0) {
+	/* Hack to convert arguments to numbers for backwards compat. Ugh. */
+	scriptbuf = rstrscat(NULL,
+		arg1 >= 0 ? "arg[2] = tonumber(arg[2]);" : "",
+		arg2 >= 0 ? "arg[3] = tonumber(arg[3]);" : "",
+		script->body, NULL);
     }
-    if (arg2 >= 0) {
-	rpmluavSetValueNum(var, arg2);
-	rpmluaSetVar(lua, var);
-    }
-    rpmluaPop(lua);
 
     /* Lua scripts can change our cwd and umask, save and restore */
-    /* XXX TODO: use cwd from chroot state to save unnecessary open here */
     cwd = open(".", O_RDONLY);
     if (cwd != -1) {
 	mode_t oldmask = umask(0);
 	umask(oldmask);
-	pid_t pid = getpid();
 
-	if (chdir("/") == 0 && rpmluaRunScript(lua, script, sname) == 0) {
+	lua_pushstring(L, "RPM_PACKAGE_RPMVERSION");
+	lua_pushstring(L, script->rpmver);
+	lua_settable(L, LUA_REGISTRYINDEX);
+
+	if (chdir("/") == 0 &&
+		rpmluaRunScript(lua, scriptbuf, script->descr, NULL, *argvp) == 0) {
 	    rc = RPMRC_OK;
 	}
-	if (pid != getpid()) {
-	    /* Terminate child process forked in lua scriptlet */
-	    rpmlog(RPMLOG_ERR, _("No exec() called after fork() in lua scriptlet\n"));
-	    _exit(EXIT_FAILURE);
-	}
+
+	lua_getfield(L, LUA_REGISTRYINDEX, "RPM_PACKAGE_RPMVERSION");
+	lua_pushnil(L);
+	lua_settable(L, LUA_REGISTRYINDEX);
+
 	/* This failing would be fatal, return something different for it... */
 	if (fchdir(cwd)) {
 	    rpmlog(RPMLOG_ERR, _("Unable to restore current directory: %m"));
@@ -155,13 +201,19 @@ static rpmRC runLuaScript(rpmPlugins plugins, ARGV_const_t prefixes,
 	close(cwd);
 	umask(oldmask);
     }
+    if (scriptbuf != script->body)
+	free(scriptbuf);
 
-    rpmluaDelVar(lua, "arg");
-    rpmluavFree(var);
+    if (prefixes) {
+	lua_pushnil(L);
+	lua_setglobal(L, "RPM_INSTALL_PREFIX");
+    }
 
-#else
-    rpmlog(lvl, _("<lua> scriptlet support not built in\n"));
-#endif
+    if (script->nextFileFunc) {
+	lua_pushnil(L);
+	lua_setfield(L, -2, "next_file");
+	lua_pop(L, 1); /* "rpm" global */
+    }
 
     return rc;
 }
@@ -172,9 +224,19 @@ static void doScriptExec(ARGV_const_t argv, ARGV_const_t prefixes,
 			FD_t scriptFd, FD_t out)
 {
     int xx;
+    struct sigaction act;
+    sigset_t set;
+
+    /* Unmask most signals, the scripts may need them */
+    sigfillset(&set);
+    sigdelset(&set, SIGINT);
+    sigdelset(&set, SIGQUIT);
+    sigprocmask(SIG_UNBLOCK, &set, NULL);
 
     /* SIGPIPE is ignored in rpm, reset to default for the scriptlet */
-    (void) signal(SIGPIPE, SIG_DFL);
+    memset(&act, 0, sizeof(act));
+    act.sa_handler = SIG_DFL;
+    sigaction(SIGPIPE, &act, NULL);
 
     rpmSetCloseOnExec();
 
@@ -217,9 +279,6 @@ static void doScriptExec(ARGV_const_t argv, ARGV_const_t prefixes,
     }
 	
     if (chdir("/") == 0) {
-	/* XXX Don't mtrace into children. */
-	unsetenv("MALLOC_CHECK_");
-
 	xx = execv(argv[0], argv);
 	if (xx) {
 	    rpmlog(RPMLOG_ERR,
@@ -258,28 +317,27 @@ exit:
  * Run an external script.
  */
 static rpmRC runExtScript(rpmPlugins plugins, ARGV_const_t prefixes,
-		   const char *sname, rpmlogLvl lvl, FD_t scriptFd,
-		   ARGV_t * argvp, const char *script, int arg1, int arg2,
-		   scriptNextFileFunc nextFileFunc)
+		   rpmScript script, rpmlogLvl lvl, FD_t scriptFd,
+		   ARGV_t * argvp, int arg1, int arg2)
 {
     FD_t out = NULL;
     char * fn = NULL;
     pid_t pid, reaped;
     int status;
-    int inpipe[2];
+    int inpipe[2] = { -1, -1 };
     FILE *in = NULL;
     const char *line;
     char *mline = NULL;
     rpmRC rc = RPMRC_FAIL;
 
-    rpmlog(RPMLOG_DEBUG, "%s: scriptlet start\n", sname);
+    rpmlog(RPMLOG_DEBUG, "%s: scriptlet start\n", script->descr);
 
-    if (script) {
-	fn = writeScript(*argvp[0], script);
+    if (script->body) {
+	fn = writeScript(*argvp[0], script->body);
 	if (fn == NULL) {
 	    rpmlog(RPMLOG_ERR,
 		   _("Couldn't create temporary file for %s: %s\n"),
-		   sname, strerror(errno));
+		   script->descr, strerror(errno));
 	    goto exit;
 	}
 
@@ -314,18 +372,18 @@ static rpmRC runExtScript(rpmPlugins plugins, ARGV_const_t prefixes,
     }
     if (out == NULL) { 
 	rpmlog(RPMLOG_ERR, _("Couldn't duplicate file descriptor: %s: %s\n"),
-	       sname, strerror(errno));
+	       script->descr, strerror(errno));
 	goto exit;
     }
 
     pid = fork();
     if (pid == (pid_t) -1) {
 	rpmlog(RPMLOG_ERR, _("Couldn't fork %s: %s\n"),
-		sname, strerror(errno));
+		script->descr, strerror(errno));
 	goto exit;
     } else if (pid == 0) {/* Child */
 	rpmlog(RPMLOG_DEBUG, "%s: execv(%s) pid %d\n",
-	       sname, *argvp[0], (unsigned)getpid());
+	       script->descr, *argvp[0], (unsigned)getpid());
 
 	fclose(in);
 	dup2(inpipe[0], STDIN_FILENO);
@@ -340,7 +398,8 @@ static rpmRC runExtScript(rpmPlugins plugins, ARGV_const_t prefixes,
     close(inpipe[0]);
     inpipe[0] = 0;
 
-    if (nextFileFunc->func) {
+    if (script->nextFileFunc) {
+	scriptNextFileFunc nextFileFunc = script->nextFileFunc;
 	while ((line = nextFileFunc->func(nextFileFunc->param)) != NULL) {
 	    size_t size = strlen(line);
 	    size_t ret_size;
@@ -368,18 +427,18 @@ static rpmRC runExtScript(rpmPlugins plugins, ARGV_const_t prefixes,
     } while (reaped == -1 && errno == EINTR);
 
     rpmlog(RPMLOG_DEBUG, "%s: waitpid(%d) rc %d status %x\n",
-	   sname, (unsigned)pid, (unsigned)reaped, status);
+	   script->descr, (unsigned)pid, (unsigned)reaped, status);
 
     if (reaped < 0) {
 	rpmlog(lvl, _("%s scriptlet failed, waitpid(%d) rc %d: %s\n"),
-		 sname, pid, reaped, strerror(errno));
+		 script->descr, pid, reaped, strerror(errno));
     } else if (!WIFEXITED(status) || WEXITSTATUS(status)) {
       	if (WIFSIGNALED(status)) {
 	    rpmlog(lvl, _("%s scriptlet failed, signal %d\n"),
-                   sname, WTERMSIG(status));
+                   script->descr, WTERMSIG(status));
 	} else {
 	    rpmlog(lvl, _("%s scriptlet failed, exit status %d\n"),
-		   sname, WEXITSTATUS(status));
+		   script->descr, WEXITSTATUS(status));
 	}
     } else {
 	/* if we get this far we're clear */
@@ -432,9 +491,9 @@ rpmRC rpmScriptRun(rpmScript script, int arg1, int arg2, FD_t scriptFd,
 
     if (rc != RPMRC_FAIL) {
 	if (script_type & RPMSCRIPTLET_EXEC) {
-	    rc = runExtScript(plugins, prefixes, script->descr, lvl, scriptFd, &args, script->body, arg1, arg2, &script->nextFileFunc);
+	    rc = runExtScript(plugins, prefixes, script, lvl, scriptFd, &args, arg1, arg2);
 	} else {
-	    rc = runLuaScript(plugins, prefixes, script->descr, lvl, scriptFd, &args, script->body, arg1, arg2, &script->nextFileFunc);
+	    rc = runLuaScript(plugins, prefixes, script, lvl, scriptFd, &args, arg1, arg2);
 	}
     }
 
@@ -475,11 +534,13 @@ static rpmScript rpmScriptNew(Header h, rpmTagVal tag, const char *body,
 			      rpmscriptFlags flags, const char *prefix)
 {
     char *nevra = headerGetAsString(h, RPMTAG_NEVRA);
-    rpmScript script = xcalloc(1, sizeof(*script));
+    rpmScript script = (rpmScript)xcalloc(1, sizeof(*script));
     script->tag = tag;
     script->type = getScriptType(tag);
     script->flags = getDefFlags(tag) | flags;
     script->body = (body != NULL) ? xstrdup(body) : NULL;
+    script->rpmver = headerGetAsString(h, RPMTAG_RPMVERSION);
+    script->chroot = 1;
     rasprintf(&script->descr, "%%%s%s(%s)", prefix, tag2sln(tag), nevra);
 
     /* macros need to be expanded before possible queryformat */
@@ -495,18 +556,16 @@ static rpmScript rpmScriptNew(Header h, rpmTagVal tag, const char *body,
 	script->body = b;
     }
 
-    script->nextFileFunc.func = NULL;
-    script->nextFileFunc.param = NULL;
-
     free(nevra);
     return script;
 }
 
-void rpmScriptSetNextFileFunc(rpmScript script, char *(*func)(void *),
+void rpmScriptSetNextFileFunc(rpmScript script, nextfilefunc func,
 			    void *param)
 {
-    script->nextFileFunc.func = func;
-    script->nextFileFunc.param = param;
+    script->nextFileFunc = (struct scriptNextFileFunc_s *)xmalloc(sizeof(*script->nextFileFunc));
+    script->nextFileFunc->func = func;
+    script->nextFileFunc->param = param;
 }
 
 rpmTagVal triggerDsTag(rpmscriptTriggerModes tm)
@@ -591,6 +650,8 @@ rpmScript rpmScriptFromTriggerTag(Header h, rpmTagVal triggerTag,
 	    headerGet(h, RPMTAG_TRANSFILETRIGGERSCRIPTFLAGS, &tflags, hgflags);
 	    prefix = "transfile";
 	    break;
+	default:
+	    return NULL;
     }
 
     if (rpmtdSetIndex(&tscripts, ix) >= 0 && rpmtdSetIndex(&tprogs, ix) >= 0) {
@@ -604,7 +665,7 @@ rpmScript rpmScriptFromTriggerTag(Header h, rpmTagVal triggerTag,
 				rpmtdGetString(&tscripts), sflags, prefix);
 
 	/* hack up a hge-style NULL-terminated array */
-	script->args = xmalloc(2 * sizeof(*script->args) + strlen(prog) + 1);
+	script->args = (char **)xmalloc(2 * sizeof(*script->args) + strlen(prog) + 1);
 	script->args[0] = (char *)(script->args + 2);
 	script->args[1] = NULL;
 	strcpy(script->args[0], prog);
@@ -617,6 +678,19 @@ rpmScript rpmScriptFromTriggerTag(Header h, rpmTagVal triggerTag,
     return script;
 }
 
+rpmScript rpmScriptFromArgv(Header h, rpmTagVal scriptTag, ARGV_t argv, rpmscriptFlags flags, int chroot)
+{
+    rpmScript script = NULL;
+
+    if (h && argv) {
+	char *body = argvJoin(argv, " ");
+	script = rpmScriptNew(h, scriptTag, body, flags, "");
+	script->chroot = chroot;
+	free(body);
+    }
+    return script;
+}
+
 rpmScript rpmScriptFromTag(Header h, rpmTagVal scriptTag)
 {
     rpmScript script = NULL;
@@ -630,7 +704,7 @@ rpmScript rpmScriptFromTag(Header h, rpmTagVal scriptTag)
 			      headerGetNumber(h, getFlagTag(scriptTag)), "");
 
 	if (headerGet(h, progTag, &prog, (HEADERGET_ALLOC|HEADERGET_ARGV))) {
-	    script->args = prog.data;
+	    script->args = (char **)prog.data;
 	}
     }
     return script;
@@ -642,6 +716,8 @@ rpmScript rpmScriptFree(rpmScript script)
 	free(script->args);
 	free(script->body);
 	free(script->descr);
+	free(script->nextFileFunc);
+	free(script->rpmver);
 	free(script);
     }
     return NULL;
